from dtm import DTM
from mtm import MTM
from _helpers import (
    read_file,
    get_dtm_configuration,
    get_mtm_configuration,
    get_dtm_transition_function,
    validate_dtm_configuration,
    validate_mtm_configuration,
    validate_dtm_transitions,
    validate_mtm_transitions,
    get_mtm_transition_function,
)
from typing import Optional
from sys import stderr


def load_dtm_from_file(file_path: str) -> Optional[DTM]:
    """
    Loads a DTM from a file if it is valid.
    Returns None if the configuration is invalid and prints the error to stderr.

    Args:
        file_path (str): Path to the file with configuration.

    Returns:
        Optional[DTM]: DTM if the configuration is valid, None otherwise.
    """
    definition = read_file(file_path)

    config_err = validate_dtm_configuration(definition)
    transitions_err = validate_dtm_transitions(definition)
    if config_err or transitions_err:
        print(config_err or transitions_err, file=stderr)
        return None

    init, acc, rej, abc = get_dtm_configuration(definition)
    transitions = get_dtm_transition_function(definition)

    return DTM(
        states={*transitions.keys()},
        input_alphabet=abc,
        acc_state=acc,
        rej_state=rej,
        initial_state=init,
        transitions=transitions,
    )


def load_mtm_from_file(file_path: str) -> Optional[MTM]:
    """
    Loads a MTM from a file if it is valid.
    Returns None if the configuration is invalid and prints the error to stderr.

    Args:
        file_path (str): Path to the file with configuration.

    Returns:
        Optional[MTM]: MTM if the configuration is valid, None otherwise.
    """
    definition = read_file(file_path)

    config_err = validate_mtm_configuration(definition)
    transitions_err = validate_mtm_transitions(definition)
    if config_err or transitions_err:
        print(config_err or transitions_err, file=stderr)
        return None

    init, acc, rej, abc, tape_count = get_mtm_configuration(definition)
    transitions = get_mtm_transition_function(definition)

    return MTM(
        states={*transitions.keys()},
        input_alphabet=abc,
        acc_state=acc,
        rej_state=rej,
        initial_state=init,
        tape_count=tape_count,
        transitions=transitions,
    )


if __name__ == "__main__":
    machine = load_dtm_from_file("./test.txt")
    machine.write_to_tape("abbabba")
    machine.simulate(step_by_step=True)



from hypothesis import given
from sys import path
from generation import acc_palindromes, rej_palindromes

path.append("../src/ib110hw")
from turing.dtm import DTM
from turing.tape import Direction

PALINDROME_MACHINE: DTM = DTM(
    states={
        "init",
        "mark",
        "gotoEndA",
        "checkA",
        "gotoEndB",
        "checkB",
        "accept",
        "reject",
    },
    input_alphabet={"a", "b"},
    acc_state="accept",
    rej_state="reject",
    initial_state="init",
    transitions={
        "init": {
            ">": ("mark", ">", Direction.RIGHT),
        },
        "mark": {
            "a": ("foundA", "X", Direction.RIGHT),
            "b": ("foundB", "X", Direction.RIGHT),
            "X": ("accept", "X", Direction.STAY),
            "": ("accept", "", Direction.STAY),
        },
        "foundA": {
            "a": ("foundA", "a", Direction.RIGHT),
            "b": ("foundA", "b", Direction.RIGHT),
            "X": ("checkA", "X", Direction.LEFT),
            "": ("checkA", "", Direction.LEFT),
        },
        "checkA": {
            "a": ("back", "X", Direction.LEFT),
            "b": ("reject", "b", Direction.STAY),
            "X": ("accept", "X", Direction.STAY),
        },
        "foundB": {
            "a": ("foundB", "a", Direction.RIGHT),
            "b": ("foundB", "b", Direction.RIGHT),
            "X": ("checkB", "X", Direction.LEFT),
            "": ("checkB", "", Direction.LEFT),
        },
        "checkB": {
            "a": ("reject", "a", Direction.STAY),
            "b": ("back", "X", Direction.LEFT),
            "X": ("accept", "X", Direction.STAY),
        },
        "back": {
            "a": ("back", "a", Direction.LEFT),
            "b": ("back", "b", Direction.LEFT),
            "X": ("mark", "X", Direction.RIGHT),
        },
    },
)

PALINDROME_MACHINE.max_steps = 10000


@given(acc_palindromes())
def test_simulate_acc(input_str: str):
    PALINDROME_MACHINE.write_to_tape(input_str)
    assert PALINDROME_MACHINE.simulate(to_console=False)
    PALINDROME_MACHINE.clear_tape()


@given(rej_palindromes())
def test_simulate_rej(input_str: str):
    PALINDROME_MACHINE.write_to_tape(input_str)
    assert not PALINDROME_MACHINE.simulate(to_console=False)
    PALINDROME_MACHINE.clear_tape()



from .dtm import DTM
from .tape import Direction

if __name__ == "__main__":
    transitions = {
        "init": {
            ">": ("findA", ">", Direction.RIGHT),
        },
        "findA": {
            "a": ("findB", "a", Direction.RIGHT),
            "b": ("findA", "b", Direction.RIGHT),
            "c": ("findA", "c", Direction.RIGHT),
        },
        "findB": {
            "a": ("findB", "a", Direction.RIGHT),
            "b": ("markC", "b", Direction.RIGHT),
            "c": ("findA", "c", Direction.RIGHT),
        },
        "markC": {
            "a": ("shiftA", "X", Direction.RIGHT),
            "b": ("shiftB", "X", Direction.RIGHT),
            "c": ("shiftC", "X", Direction.RIGHT),
            "": ("findA", "c", Direction.RIGHT),
        },
        "shiftA": {
            "a": ("shiftA", "a", Direction.RIGHT),
            "b": ("shiftB", "a", Direction.RIGHT),
            "c": ("shiftC", "a", Direction.RIGHT),
            "": ("goBack", "a", Direction.LEFT),
        },
        "shiftB": {
            "a": ("shiftA", "b", Direction.RIGHT),
            "b": ("shiftB", "b", Direction.RIGHT),
            "c": ("shiftC", "b", Direction.RIGHT),
            "": ("goBack", "b", Direction.LEFT),
        },
        "shiftC": {
            "a": ("shiftA", "c", Direction.RIGHT),
            "b": ("shiftB", "c", Direction.RIGHT),
            "c": ("shiftC", "c", Direction.RIGHT),
            "": ("goBack", "c", Direction.LEFT),
        },
        "goBack": {
            "a": ("goBack", "a", Direction.LEFT),
            "b": ("goBack", "b", Direction.LEFT),
            "c": ("goBack", "c", Direction.LEFT),
            "X": ("findA", "c", Direction.RIGHT),
        },
    }

    machine = DTM(
        states={*transitions.keys()},
        input_alphabet={"a", "b", "c"},
        transitions=transitions,
    )

    machine.write_to_tape("abba")
    machine.simulate(step_by_step=True)



from copy import deepcopy
from time import sleep
from typing import Dict, List, Optional, Set, Tuple, IO

from base import BaseTuringMachine, MAX_STEPS_ERROR_MSG
from tape import Direction, Tape, START_SYMBOL
from _helpers import clear_console, close_file, mtm_config_to_md, get_step_direction

Symbols = Tuple[str, ...]
Directions = Tuple[Direction, ...]

MTMRule = Tuple[str, Symbols, Directions]
MTMRules = Dict[Symbols, MTMRule]
MTMTransitions = Dict[str, MTMRules]


class MTM(BaseTuringMachine):
    """
    Represents a Multi-tape Turing Machine
    """

    def __init__(
        self,
        states: Set[str] = None,
        input_alphabet: Set[str] = None,
        acc_state: str = "accept",
        rej_state: str = "reject",
        transitions: MTMTransitions = None,
        tape_count: int = 2,
        tapes: List[Tape] = None,
        initial_state: str = "init",
        start_symbol: str = START_SYMBOL,
    ):
        super().__init__(
            states,
            input_alphabet,
            acc_state,
            rej_state,
            initial_state,
            start_symbol,
        )
        self.transitions = transitions if transitions is not None else {}
        self.tapes = tapes or [deepcopy(Tape()) for _ in range(tape_count)]
        self.tape_count = tape_count or len(tapes)

    def get_transition(self, state: str, read: Symbols) -> Optional[MTMRule]:
        """
        Returns the transition based on the current state and read symbols.

        Args:
            state (str): Current state.
            read (Symbols): Current read symbols.

        Returns:
            Optional[MTMRule]: Transition based on the provided params if exists, None otherwise.
        """
        return self.transitions.get(state, {}).get(read, None)

    def write_to_tape(self, input_str: str, index: int = 0) -> None:
        """
        Writes the provided string on the tape on the index.

        Args:
            input_str (str): String to be written on the tape.
            index (int, optional): Index of the tape to be updated. Defaults to 0.
        """
        self.tapes[index].write(self.start_symbol + input_str)

    def clear_tape(self, index: int) -> None:
        """
        Clears the tape on the index.

        Args:
            index (int): Index of the tape.
        """
        self.tapes[index].clear()

    def clear_tapes(self) -> None:
        """
        Clears all the tapes
        """
        for tape in self.tapes:
            tape.clear()

    def read_tape(self, index: int = 0) -> str:
        """
        Reads contents of a tape by index.

        Returns:
            str: String written on the tape on index.
        """
        return self.tapes[index].read()

    def get_current_symbols(self) -> Symbols:
        """
        Gets the current symbols read by tape heads.

        Returns:
            Symbols: Symbols read by the tape heads.
        """
        return tuple((tape.current.value for tape in self.tapes))

    def simulate(
        self,
        to_console: bool = True,
        to_file: bool = False,
        path: str = "simulation.md",
        delay: float = 0.5,
        step_by_step: bool = False,
    ) -> bool:
        """Simulates the machine on its current tape configuration.

        Args:
            to_console (bool, optional): Set to False if you only want to see the result. Defaults to True.
            to_file (bool, optional): Set to True if you want to save every step to the file. Defaults to False.
            path (str, optional): Path to the .md file with the step history. Defaults to "simulation.md".
            delay (float, optional): The delay (s) between each step when printing to console. Defaults to 0.5.
            step_by_step (bool, optional): Set to True if you want the simulation wait after each step. Defaults to False.

        Returns:
            bool: False if the machine rejects the word or exceeds the 'max_steps' value, True otherwise.
        """
        state: str = self.initial_state
        steps: int = 1
        rule: Optional[MTMRule] = None
        output_file: Optional[IO] = open(path, "w") if to_file else None

        # key pressed by user when the 'step_by_step' is enabled, can be <left|right|esc>
        pressed_key: str = "right"
        prev_states: List[Tuple[str, List[Tape]]] = []

        def get_rule_string() -> str:
            """
            Parse the current configuration to string shown to user.
            """
            row = self.get_transition(state, self.get_current_symbols())
            step_index = steps if row else steps - 1
            next_step = f"{state}, {self.get_current_symbols()}"

            return f"{step_index}. ({next_step}) -> {row}"

        def write_machine_configuration() -> None:
            """
            Write the configuration string to all enabled outputs.
            """
            rule_str = get_rule_string()

            if output_file:
                output_file.write(mtm_config_to_md(self.tapes, rule_str))

            if to_console or step_by_step:
                clear_console()
                print(rule_str, "\n\n")

                for i, t in enumerate(self.tapes):
                    print(f"Tape {i}\n{t}")

                sleep(0 if step_by_step else delay)

        def go_forward() -> None:
            nonlocal rule, steps
            rule = self.get_transition(state, self.get_current_symbols())
            prev_states.append((state, deepcopy(self.tapes)))
            steps += 1

        def go_back() -> None:
            if not prev_states:
                return

            nonlocal steps, state
            steps -= 1
            state, self.tapes = prev_states.pop()

        while steps <= (self.max_steps + 1):
            write_machine_configuration()

            if state == self.acc_state:
                close_file(output_file)
                return True

            if step_by_step:
                pressed_key = get_step_direction()

            if pressed_key == "esc":
                close_file(output_file)
                print("Canceled the computation.")
                return False

            if pressed_key == "left":
                go_back()
                continue

            go_forward()

            if not rule or rule[0] == self.rej_state:
                close_file(output_file)
                return False

            state, write, directions = rule
            for direction, tape, symbol in zip(directions, self.tapes, write):
                tape.write_symbol(symbol)
                tape.move(direction)

        close_file(output_file)
        print(MAX_STEPS_ERROR_MSG.format(self.max_steps))

        return False


if __name__ == "__main__":
    transitions: MTMTransitions = {
        "init": {
            (">", ""): ("copy", (">", ""), (Direction.RIGHT, Direction.STAY))
        },
        "copy": {
            ("a", ""): ("copy", ("a", "a"), (Direction.RIGHT, Direction.RIGHT)),
            ("b", ""): ("copy", ("b", "b"), (Direction.RIGHT, Direction.RIGHT)),
            ("", ""): ("goToStart", ("", ""), (Direction.LEFT, Direction.STAY)),
        },
        "goToStart": {
            ("a", ""): ("goToStart", ("a", ""), (Direction.LEFT, Direction.STAY)),
            ("b", ""): ("goToStart", ("b", ""), (Direction.LEFT, Direction.STAY)),
            (">", ""): ("check", (">", ""), (Direction.RIGHT, Direction.LEFT))
        },
        "check": {
            ("a", "a"): ("check", ("a", "a"), (Direction.RIGHT, Direction.LEFT)),
            ("b", "b"): ("check", ("b", "b"), (Direction.RIGHT, Direction.LEFT)),
            ("", ""): ("accept", ("", ""), (Direction.STAY, Direction.STAY)),
            ("a", "b"): ("reject", ("a", "b"), (Direction.STAY, Direction.STAY)),
            ("b", "a"): ("reject", ("b", "a"), (Direction.STAY, Direction.STAY)),
        }
    }

    machine: MTM = MTM(
        states={"init", "goToEnd", "goToStart", "check", "accept", "reject"},
        initial_state="init",
        input_alphabet={"a", "b"},
        acc_state="accept",
        rej_state="reject",
        transitions=transitions)

    machine.write_to_tape("aabbabbaa")
    print(machine.simulate(step_by_step=True))



from copy import deepcopy
from time import sleep
from typing import Dict, Optional, Set, Tuple, List, IO

from base import BaseTuringMachine, MAX_STEPS_ERROR_MSG
from tape import Direction, Tape, START_SYMBOL
from _helpers import dtm_config_to_md, clear_console, close_file, get_step_direction

DTMRule = Tuple[str, str, Direction]
DTMRules = Dict[str, DTMRule]
DTMTransitions = Dict[str, DTMRules]


class DTM(BaseTuringMachine):
    """
    Represents a Deterministic Turing Machine
    """

    def __init__(
        self,
        states: Set[str] = None,
        input_alphabet: Set[str] = None,
        acc_state: str = "accept",
        rej_states: str = "reject",
        transitions: DTMTransitions = None,
        tape: Tape = None,
        initial_state: str = "init",
        start_symbol: str = START_SYMBOL,
    ) -> None:
        super().__init__(
            states,
            input_alphabet,
            acc_state,
            rej_states,
            initial_state,
            start_symbol,
        )
        self.transitions = transitions if transitions is not None else {}
        self.tape = tape or Tape()

    def get_transition(self, state: str, read: str) -> Optional[DTMRule]:
        """
        Gets the transition based on the provided state and read symbol.

        Args:
            state (str): Current state of the DTM.
            read (str): Symbol read by the tape head.

        Returns:
            DTMRule: Transition based on the provided parms if exists, None otherwise.
        """
        return self.transitions.get(state, {}).get(read, None)

    def write_to_tape(self, input_str: str) -> None:
        """
        Writes the provided string on the tape.

        Args:
            input_str (str): String to be written on the tape.
        """
        self.tape.write(self.start_symbol + input_str)

    def clear_tape(self) -> None:
        """
        Clears the contents of the tape.
        """
        self.tape.clear()

    def read_tape(self) -> str:
        """
        Reads contents of the tape.

        Returns:
            str: String written on the tape.
        """
        return self.tape.read()

    def simulate(
        self,
        to_console: bool = True,
        to_file: bool = False,
        path: str = "simulation.md",
        delay: float = 0.5,
        step_by_step: bool = False,
    ) -> bool:
        """
        Simulates the machine on its current tape configuration.

        Args:
            to_console (bool, optional): Set to False if you only want to see the result. Defaults to True.
            to_file (bool, optional): Set to True if you want to save every step to the file. Defaults to False.
            path (str, optional): Path to the .md file with the step history. Defaults to "simulation.md".
            delay (float, optional): The delay (s) between each step when printing to console. Defaults to 0.5.
            step_by_step (bool, optional): Set to True if you want the simulation wait after each step. Defaults to False.

        Returns:
            bool: False if the machine rejects the word or exceeds the 'max_steps' value, True otherwise.
        """
        state: str = self.initial_state
        steps: int = 0
        rule: Optional[DTMRule] = None
        output_file: Optional[IO] = open(path, "w") if to_file else None

        # key pressed by user when the 'step_by_step' is enabled, can be <left|right|esc>
        pressed_key: str = "right"

        # remember the previous state and tape configuration to be able to go back and forth
        prev_states: List[Tuple[str, Tape]] = []

        def get_rule_string() -> str:
            """
            Parse the current configuration to string shown to user.
            """
            row = self.get_transition(state, self.tape.current.value)
            step_index = steps if row else steps - 1
            next_step = f"{state}, '{self.tape.current.value}'"

            return f"{step_index}. ({next_step}) -> {row}\n"

        def write_machine_configuration() -> None:
            """
            Write the configuration string to all enabled outputs.
            """
            rule_str = get_rule_string()

            if output_file:
                output_file.write(dtm_config_to_md(self.tape, rule_str))

            if to_console or step_by_step:
                clear_console()
                print(f"{rule_str}\n{self.tape}")
                sleep(0 if step_by_step else delay)

        def go_forward() -> None:
            """
            Goes one step forward in the computation.
            """
            nonlocal steps, rule, prev_states
            rule = self.get_transition(state, self.tape.current.value)
            prev_states.append((state, deepcopy(self.tape)))
            steps += 1

        def go_back() -> None:
            """
            Goes one step back in the computation.
            Does nothing if there is no previous step.
            """
            if not prev_states:
                return

            nonlocal steps, state
            steps -= 1
            state, self.tape = prev_states.pop()

        while steps <= (self.max_steps + 1):
            write_machine_configuration()

            if state == self.acc_state:
                close_file(output_file)
                return True

            if step_by_step:
                pressed_key = get_step_direction()

            if pressed_key == "esc":
                close_file(output_file)
                print("Canceled the computation.")
                return False

            if pressed_key == "left":
                go_back()
                continue

            go_forward()

            if not rule or rule[0] == self.rej_state:
                close_file(output_file)
                return False

            state, write, direction = rule
            self.tape.write_symbol(write)
            self.tape.move(direction)

        close_file(output_file)
        print(MAX_STEPS_ERROR_MSG.format(self.max_steps))

        return False


if __name__ == "__main__":
    transitions = {
        "init": {
            ">": ("findA", ">", Direction.R),
        },
        "findA": {
            "a": ("findB", "a", Direction.R),
            "b": ("findA", "b", Direction.R),
            "c": ("findA", "c", Direction.R),
        },
        "findB": {
            "a": ("findB", "a", Direction.R),
            "b": ("markC", "b", Direction.R),
            "c": ("findA", "c", Direction.R),
        },
        "markC": {
            "a": ("shiftA", "X", Direction.R),
            "b": ("shiftB", "X", Direction.R),
            "c": ("shiftC", "X", Direction.R),
            "": ("findA", "c", Direction.R),
        },
        "shiftA": {
            "a": ("shiftA", "a", Direction.R),
            "b": ("shiftB", "a", Direction.R),
            "c": ("shiftC", "a", Direction.R),
            "": ("goBack", "a", Direction.L),
        },
        "shiftB": {
            "a": ("shiftA", "b", Direction.R),
            "b": ("shiftB", "b", Direction.R),
            "c": ("shiftC", "b", Direction.R),
            "": ("goBack", "b", Direction.L),
        },
        "shiftC": {
            "a": ("shiftA", "c", Direction.R),
            "b": ("shiftB", "c", Direction.R),
            "c": ("shiftC", "c", Direction.R),
            "": ("goBack", "c", Direction.L),
        },
        "goBack": {
            "a": ("goBack", "a", Direction.L),
            "b": ("goBack", "b", Direction.L),
            "c": ("goBack", "c", Direction.L),
            "X": ("findA", "c", Direction.R),
        },
    }

    machine = DTM(
        states={*transitions.keys()},
        input_alphabet={"a", "b", "c"},
        transitions=transitions,
    )

    machine.write_to_tape("abba")
    print(machine.simulate(step_by_step=True))
